/* Based on alpha/db_trace.c and FreeBSD: ia64/db_interface.c, ia64/db_unwind.c */

#include "opt_ddb.h"

#include <sys/param.h>
#include <sys/proc.h>
#include <sys/systm.h>

#include <machine/cpufunc.h>
#include <machine/md_var.h>
#include <machine/db_machdep.h>
#include <machine/unwind.h>

#include <ddb/db_sym.h>
#include <ddb/db_access.h>
#include <ddb/db_output.h>
#include <ddb/db_variables.h>
#include <ddb/db_interface.h>

#include <ia64/unwind/decode.h>
#include <ia64/unwind/stackframe.h>


void
initunwindframe(struct unwind_frame *uwf, struct trapframe *tf);
void
rewindframe(struct unwind_frame *uwf, db_addr_t ip);

void
db_stack_trace_print(db_expr_t addr, boolean_t have_addr, db_expr_t count,
		     const char *modif, void (*pr)(const char *, ...))
{
	char c;
	const char *cp = modif;
	boolean_t trace_thread = FALSE;
	boolean_t trace_user = FALSE;
	struct trapframe *tf;
	struct unwind_frame current_frame;
	db_addr_t ip;
	const char *name;
	db_sym_t sym;
	db_expr_t offset;

	while ((c = *cp++) != 0) {
		trace_thread |= c == 't';
		trace_user |= c == 'u';
	}

	if (trace_user) {
		(*pr)("User-space stack tracing not implemented yet. \n");
		return;
	}
	if (!have_addr) {
		(*pr)("--Kernel Call Trace-- \n");

		tf = DDB_REGS;
		ip = tf->tf_special.iip + ((tf->tf_special.psr >> 41) & 3);

		initunwindframe(&current_frame, tf);
		patchunwindframe(&current_frame, ip - kernstart, kernstart);

		/* Get into unwind loop. */

		int rubbish = 0;
		while(ip) {
			sym = db_search_symbol(ip, DB_STGY_ANY, &offset);
			db_symbol_values(sym, &name, NULL);
			(*pr)("%s(...)\n", name);

			ip = current_frame.rp;

			if(!ip) break;

			rewindframe(&current_frame, ip);
			rubbish++;
		}

		return;


	} else (*pr) ("Unwind from arbitrary addresses unimplemented. \n");
		

		if (trace_thread) {
			(*pr)("trace by pid unimplemented. \n");
			return;
		}
		else {
			(*pr)("trace from arbitrary trap frame address unimplemented. \n");
		}

}

extern db_addr_t ia64_unwindtab;
extern vsize_t ia64_unwindtablen;


/* Generates initial unwind frame context based on the contents
 * of the trap frame, by consulting the Unwind library 
 * staterecord. If a register is of type enum UNSAVED, we fetch
 * the live value of the register from the trapframe.
 */

void
initunwindframe(struct unwind_frame *uwf, struct trapframe *tf)
		 
{

	uwf->rp = tf->tf_special.rp;

	/* ndirty = bsp - bspstore: , not the same as the definition in the spec.
	 * Gave me hell for a day!
	 * see: ia64/exception.S: exception_save_restore: */

	uwf->bsp = tf->tf_special.bspstore + tf->tf_special.ndirty;
	uwf->bsp = ia64_bsp_adjust_ret(uwf->bsp, IA64_CFM_SOF(tf->tf_special.cfm));
	uwf->pfs = tf->tf_special.pfs;
	uwf->sp = tf->tf_special.sp;


}
		


/* Single step the frame backward. 
 * Assumes unwind_frame is setup already.
 */

void
rewindframe(struct unwind_frame *uwf, db_addr_t ip)
{
/* XXX: Check for a stack switch */

#if 0
	printf("XXX: uwf->bsp before = %lx \n", uwf->bsp);
	printf("XXX: SOL(pfs) = %lu \n", IA64_CFM_SOL(uwf->pfs));
#endif
	uwf->bsp = ia64_bsp_adjust_ret(uwf->bsp, IA64_CFM_SOL(uwf->pfs));
#if 0
	printf("XXX: uwf->bsp = %lx \n", uwf->bsp);
#endif

	/* Stomp on rp and pfs 
	 */
	patchunwindframe(uwf, ip - kernstart, kernstart);

}
